package cn.com.qmth.scancloud.multithread.consumer;

import cn.com.qmth.scancloud.bean.SchoolSession;
import cn.com.qmth.scancloud.bean.answersave.ArrayResult;
import cn.com.qmth.scancloud.bean.answersave.StringResult;
import cn.com.qmth.scancloud.client.MarkingcloudApiClient;
import cn.com.qmth.scancloud.entity.*;
import cn.com.qmth.scancloud.enums.LockType;
import cn.com.qmth.scancloud.enums.ScanStatus;
import cn.com.qmth.scancloud.enums.UploadStatus;
import cn.com.qmth.scancloud.model.DataUploadDto;
import cn.com.qmth.scancloud.model.DataUploadResponse;
import cn.com.qmth.scancloud.model.FileUploadResponse;
import cn.com.qmth.scancloud.multithread.BatchConsumer;
import cn.com.qmth.scancloud.service.*;
import cn.com.qmth.scancloud.vo.StudentUploadVo;
import com.qmth.boot.core.concurrent.service.ConcurrentService;
import com.qmth.boot.core.exception.StatusException;
import com.qmth.boot.core.fss.store.FileStore;
import com.qmth.boot.core.retrofit.utils.SignatureInfo;
import com.qmth.boot.core.retrofit.utils.UploadFile;
import com.qmth.boot.tools.io.IOUtils;
import com.qmth.boot.tools.signature.SignatureType;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

@Scope("prototype")
@Service
public class UploadConsumer extends BatchConsumer<StudentUploadVo> {

    private static final Logger log = LoggerFactory.getLogger(UploadConsumer.class);

    @Autowired
    private StudentService studentService;

    @Autowired
    private ConcurrentService concurrentService;

    @Autowired
    private MarkingcloudApiClient markingcloudApiClient;

    @Autowired
    private BatchPaperService batchPaperService;

    @Autowired
    private PaperPageService paperPageService;

    @Autowired
    private StudentPaperService studentPaperService;

    @Autowired
    private FilePropertyService filePropertyService;

    @Autowired
    private FileStore fileStore;

    @Autowired
    private ExamService examService;

    @Autowired
    private SessionService sessionService;

    @Transactional
    @Override
    public void consume(StudentUploadVo s) {
        SchoolSession ss = sessionService.getSchoolSession(s.getSchoolId());
        if (ss == null) {
            return;
        }
        SignatureInfo signatureInfo = new SignatureInfo(SignatureType.TOKEN, ss.getAccount(),
                ss.getMarkingCloudToken());
        StudentEntity se = studentService.getById(s.getId());
        if (se == null) {
            return;
        }
        concurrentService.getReadWriteLock(LockType.STUDENT + "-" + s.getId()).writeLock().lock();
        try {
            if (se.getFileUploadStatus() == UploadStatus.WAITING_UPLOAD
                    || se.getFileUploadStatus() == UploadStatus.ERROR) {
                log.info("开始上传文件，examId={}, examNumber={}", s.getExamId(), s.getExamNumber());
                try {
                    uploadFile(signatureInfo, se);
                    studentService.updateFileUploadStatus(se.getId(), UploadStatus.UPLOADED);
                } catch (Exception e) {
                    log.error("文件上传出错，examId=" + s.getExamId() + ", examNumber=" + s.getExamNumber(), e);
                    studentService.updateFileUploadStatus(se.getId(), UploadStatus.ERROR);
                    //文件上传出错直接返回
                    return;
                }
            }
            if (se.getDataUploadStatus() == UploadStatus.WAITING_UPLOAD
                    || se.getDataUploadStatus() == UploadStatus.ERROR) {
                log.info("开始上传数据，examId={}, examNumber={}", s.getExamId(), s.getExamNumber());
                try {
                    uploadData(signatureInfo, se);
                    studentService.updateDataUploadStatus(se.getId(), UploadStatus.UPLOADED);
                } catch (Exception e) {
                    log.error("数据上传出错，examId=" + s.getExamId() + ", examNumber=" + s.getExamNumber(), e);
                    studentService.updateDataUploadStatus(se.getId(), UploadStatus.ERROR);
                }
            }
        } finally {
            concurrentService.getReadWriteLock(LockType.STUDENT + "-" + s.getId()).writeLock().unlock();
        }
    }

    private void uploadFile(SignatureInfo signatureInfo, StudentEntity student) throws Exception {
        List<StudentPaperEntity> sps = studentPaperService.findByStudentId(student.getId());
        if (CollectionUtils.isEmpty(sps)) {
            return;
        }
        List<String> sheets = new ArrayList<>();
        List<String> slices = new ArrayList<>();
        for (StudentPaperEntity sp : sps) {
            List<PaperPageEntity> pages = paperPageService.listByPaperId(sp.getPaperId());
            if (CollectionUtils.isNotEmpty(sps)) {
                for (PaperPageEntity page : pages) {
                    sheets.add(page.getSheetPath());
                    if (CollectionUtils.isNotEmpty(page.getSlicePath())) {
                        slices.addAll(page.getSlicePath());
                    }
                }
            }
        }
        if (CollectionUtils.isNotEmpty(sheets)) {
            int index = 0;
            for (String s : sheets) {
                index++;
                FileUploadResponse res = markingcloudApiClient
                        .sheetFileUpload(signatureInfo.toString(), student.getExamId(), student.getExamNumber(), index,
                                filePropertyService.getById(s).getMd5(),
                                UploadFile.build("file", getFileName(s), IOUtils.toByteArray(fileStore.read(s))));
                if (res == null || !res.getSuccess()) {
                    throw new StatusException(
                            "sheet上传失败,Success:false.examId=" + student.getExamId() + "，examNumber=" + student
                                    .getExamNumber());
                }
            }
        }
        if (CollectionUtils.isNotEmpty(slices)) {
            int index = 0;
            for (String s : slices) {
                index++;
                FileUploadResponse res = markingcloudApiClient
                        .sliceFileUpload(signatureInfo.toString(), student.getExamId(), student.getExamNumber(), index,
                                filePropertyService.getById(s).getMd5(),
                                UploadFile.build("file", getFileName(s), IOUtils.toByteArray(fileStore.read(s))));
                if (res == null || !res.getSuccess()) {
                    throw new StatusException(
                            "slice上传失败,Success:false.examId=" + student.getExamId() + "，examNumber=" + student
                                    .getExamNumber());
                }
            }
        }
    }
    private void disposeQuestionMark(ArrayResult ar) {
        if (ar != null && CollectionUtils.isNotEmpty(ar.getResult())) {
            for(int i=0;i<ar.getResult().size();i++) {
                ar.getResult().set(i, ar.getResult().get(i).replaceAll("\\?", ""));
            }
        }
    }
    private void uploadData(SignatureInfo signatureInfo, StudentEntity student) {
        ExamEntity exam = examService.getById(student.getExamId());
        List<StudentPaperEntity> sps = studentPaperService.findByStudentId(student.getId());
        List<String> sheets = new ArrayList<>();
        List<String> slices = new ArrayList<>();
        List<PaperPageEntity> pageList = new ArrayList<>();
        if (CollectionUtils.isNotEmpty(sps)) {
            for (StudentPaperEntity sp : sps) {
                List<PaperPageEntity> pages = paperPageService.listByPaperId(sp.getPaperId());
                if (CollectionUtils.isNotEmpty(sps)) {
                    pageList.addAll(pages);
                    for (PaperPageEntity page : pages) {
                        sheets.add(page.getSheetPath());
                        if (CollectionUtils.isNotEmpty(page.getSlicePath())) {
                            slices.addAll(page.getSlicePath());
                        }
                    }
                }
            }
        }
        //未扫描或者没有paper，需要删除考生原有上传状态
        if (student.getStatus().equals(ScanStatus.UNEXIST) || CollectionUtils.isEmpty(sheets)) {
            markingcloudApiClient.studentFileDelete(signatureInfo.toString(), exam.getId(), student.getExamNumber());
        }
        //有扫描结果，正常上传考生状态
        else {
            DataUploadDto dto = new DataUploadDto();
            dto.setSheetCount(sheets.size());
            dto.setSliceCount(slices.size());
            dto.setExamNumber(student.getExamNumber());
            dto.setManual(student.getAssigned());
            dto.setAbsent(student.getOmrAbsent());
            dto.setCardNumber(student.getCardNumber());
            List<String> answers = new ArrayList<String>();
            int pageIndex = 0;
            for (PaperPageEntity p : pageList) {
                pageIndex++;
                if (pageIndex == 1) {
                    dto.setPaperType(getPaperType(exam.getPaperTypeBarcodeContent(), p.getPaperType()));
                    BatchPaperEntity bp = batchPaperService.findByPaperId(p.getPaperId());
                    dto.setBatchCode(bp != null ? bp.getBatchId().toString() : "");
                }
                disposeQuestionMark(p.getQuestion());
                if (p.getQuestion() != null && CollectionUtils.isNotEmpty(p.getQuestion().getResult())) {
                    answers.addAll(p.getQuestion().getResult());
                }
            }
            dto.setAnswers(StringUtils.join(answers, ","));
            List<DataUploadResponse> res = markingcloudApiClient
                    .dataUpload(signatureInfo.toString(), student.getExamId(), Collections.singletonList(dto));
            if (CollectionUtils.isEmpty(res) || StringUtils.isBlank(res.get(0).getExamNumber())) {
                throw new StatusException(
                        "数据上传失败,Success:false.examId=" + student.getExamId() + "，examNumber=" + student
                                .getExamNumber());
            }
        }
    }

    private String getFileName(String path) {
        if (StringUtils.isBlank(path)) {
            throw new StatusException("文件路径有误:" + (path == null ? "" : path));
        }
        return path.substring(path.lastIndexOf("//") + 1);
    }

    private String getPaperType(List<String> list, StringResult sr) {
        String pt = sr != null ? sr.getResult() : null;
        if (StringUtils.isBlank(pt)) {
            return "#";
        }

        if (CollectionUtils.isEmpty(list)) {
            if (pt.length() > 1) {
                return "#";
            }
            pt = pt.toUpperCase();
            int pAscii = (int) pt.charAt(0);
            if (pAscii >= 65 && pAscii <= 90) {
                return pt;
            } else {
                return "#";
            }
        } else {
            int index = list.indexOf(pt);
            if (index == -1) {
                return "#";
            }
            return String.valueOf((char) (index + 65));
        }
    }

}
